/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.tca;

import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.utils.Constants.OrderingType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;

/**
 * This algorithm computes the bidirectional ring which minimizes the total 
 * ring length, using a GRASP heuristic. The cost of a link equals the euclidean 
 * distance between link end nodes. The algorithm executes GRASP iterations until 
 * the maxExecTime is reached. In each GRASP iteration, a solution is first created 
 * using a greedy-randomized approach. Then, this solution is the starting point 
 * of a local search (first-fit) heuristic, where two rings are considered neighbors 
 * when they have all but 2 bidirectional links in common. The greedy randomized 
 * approach is based on the concept of restricted candidate list (RCL). The ring 
 * starts with one single node chosen randomly. In each greedy iteration one node 
 * is added to the ring, randomly chosen from a RCL created in this iteration. 
 * The RCL contains the non-visited nodes which are closer to current node.
 * In particular, the nodes in the RCL are those which are at a distance from 
 * c_min to c_min + alpha * (c_max - c_min). c_min and c_max are the distances 
 * from last added node, to closest and furthest non-visited nodes. alpha is an 
 * algroithm parameter between 0 and 1, tuning the size of the RCL. When alpha=0, 
 * the algorithm is equal to pure greedy nearest neighbor heuristic. If alpha = 1, 
 * randomness is maximum: a non-visited node is chosen randomly.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class TCA_GRASP_TSP implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		if (N == 0) throw new Net2PlanException("The input design must have nodes");

		/* Initialize some variables */
		final double maxExecTimeSecs = Double.parseDouble(algorithmParameters.get("maxExecTimeSecs"));
		final double alpha = Double.parseDouble(algorithmParameters.get("alpha"));
		final double linkCapacities = Double.parseDouble(algorithmParameters.get("linkCapacities"));
		final long randomSeed = Long.parseLong(algorithmParameters.get("randomSeed"));
		final long algorithmEndTime = System.nanoTime() + (long) (1E9 * maxExecTimeSecs);
		Random r = new Random(randomSeed);

		/* Check input parameters */
		if ((alpha < 0) || (alpha > 1)) throw new Net2PlanException("The alpha factor must be between 0 and 1");
		if (linkCapacities < 0) throw new Net2PlanException("Link capacities must be a non-negative number");
		if (maxExecTimeSecs <= 0) throw new Net2PlanException("Algorithm running time must be a non-negative number");

		/* Compute the distances between each node pair */
		double[][] c_ij = netPlan.getNodeEuclideanDistanceMatrix().toArray();

		/* The best solution found so far (incumbent solution) is stored in these variables */
		ArrayList<Integer> best_nodeSequence = new ArrayList<Integer>();
		double best_cost = Double.MAX_VALUE;

		/* Main loop. Stop when maximum execution time is reached */
		while (System.nanoTime() < algorithmEndTime)
		{
			/******************* GREEDY-RANDOM STEP ***************************/
			Pair<ArrayList<Integer>, Double> greedySolution = computeGreedySolution(r, alpha, c_ij);

			/*** COMPUTE COST, AND CHECK IT IS A RING ***/
			checkRing(greedySolution, c_ij, "GREEDY SOLUTION");

			/******************* LOCAL SEARCH STEP ***************************/
			Pair<ArrayList<Integer>, Double> lsSolution = computeLocalSearchStep(greedySolution, c_ij);
			checkRing(lsSolution, c_ij, "LOCAL SEARCH SOLUTION");

			/* Check is the solution found improves best solution so far */
			if (lsSolution.getSecond() < best_cost)
			{
				checkRing(lsSolution, c_ij, "IMPROVEMENT!!");
				best_cost = lsSolution.getSecond();
				best_nodeSequence = (ArrayList<Integer>) lsSolution.getFirst().clone();
			}
		}

		/* Save the best solution found into the netPlan object */
		long[] nodeIdsVector = netPlan.getNodeIdsVector();
		netPlan.removeAllLinks();
		for (int cont = 0; cont < N - 1; cont++)
		{
			final int n1 = best_nodeSequence.get(cont);
			final int n2 = best_nodeSequence.get(cont + 1);
			final long nodeId1 = nodeIdsVector[n1];
			final long nodeId2 = nodeIdsVector[n2];
			netPlan.addLinkBidirectional(nodeId1, nodeId2, linkCapacities, c_ij[n1][n2], null);
		}
		
		final int firstRingN = best_nodeSequence.get(0);
		final int lastRingN = best_nodeSequence.get(best_nodeSequence.size() - 1);
		final long firstRingNodeId = nodeIdsVector[firstRingN];
		final long lastRingNodeId = nodeIdsVector[lastRingN];
		netPlan.addLinkBidirectional(firstRingNodeId, lastRingNodeId, linkCapacities, c_ij[firstRingN][lastRingN], null);

		/* Return printing the total distance of the solution */
		return "Ok! Cost : " + best_cost;
	}

	@Override
	public String getDescription()
	{
		return "This algorithm computes the bidirectional ring which minimizes the total ring length, using a GRASP heuristic. "
				+ "The cost of a link equals the euclidean distance between link end nodes. The algorithm executes GRASP iterations until the maxExecTime is reached."
				+ "In each GRASP iteration, a solution is first created using a greedy-randomized approach. Then, this solution is the starting point of a local search "
				+ "(first-fit) heuristic, where two rings are considered neighbors when they have all but 2 bidirectional links in common. "
				+ "The greedy randomized approach is based on the concept of restricted candidate list (RCL). The ring starts with one single node chosen randomly. In each greedy iteration"
				+ "one node is added to the ring, randomly chosen from a RCL created in this iteration. The RCL contains the non-visited nodes which are closer to current node."
				+ "In particular, the nodes in the RCL are those which are at a distance from c_min to c_min + alpha * (c_max - c_min). c_min and c_max are the distances "
				+ "from last added node, to closest and furthest non-visited nodes. alpha is an algroithm parameter between 0 and 1, tuning the size of the RCL. When alpha=0, the algorithm is equal to pure greedy"
				+ "nearest neighbor heuristic. If alpha = 1, randomness is maximum: a non-visited node is chosen randomly.";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new ArrayList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("alpha", "0.5", "Alpha factor for the greedy randomized step"));
		algorithmParameters.add(Triple.of("maxExecTimeSecs", "10", "Execution time of the algorithm. The algorithm will stop after this time, returning the best solution found"));
		algorithmParameters.add(Triple.of("linkCapacities", "100", "Capacities to set to the links"));
		algorithmParameters.add(Triple.of("randomSeed", "1", "Seed for the random number generator"));
		return algorithmParameters;
	}

	/**
	 * This function computes the randomized-greedy step in the GRASP. The ring 
	 * starts with a randomly chosen node. In each iteration, a restricted candidate 
	 * list is created with potential nodes to be added to the ring. These are 
	 * those non-visited nodes n, whose links (currentNode , n) are between c_min 
	 * and c_min + alpha (c_max - c_min), where c_min and c_max are the shortest 
	 * and longest distances to the non-visited nodes. Alpha is an algorithm 
	 * parameter 0...1. Then, next node is chosen randomly from the restricted 
	 * candidate list.
	 * 
	 * @since 1.0
	 */
	private Pair<ArrayList<Integer>, Double> computeGreedySolution(Random r, double alpha, double[][] c_ij)
	{
		/* Generate a greedy-randomized solution */
		final int N = c_ij.length;
		ArrayList<Integer> nodeSequence = new ArrayList<Integer>(N);
		final int initialNode = r.nextInt(N);
		nodeSequence.add(initialNode);
		HashSet<Integer> notVisitedNode = new LinkedHashSet<Integer>();
		for (int i = 0; i < N; i++)
			if (i != initialNode)
				notVisitedNode.add(i);

		double nodeSequenceCost = 0;
		while (nodeSequence.size() < N)
		{
			/* Create a list with the costs of possible next nodes  */
			final int currentNode = nodeSequence.get(nodeSequence.size() - 1);
			double[] costs = new double[notVisitedNode.size()];
			int[] nextNodes = new int[notVisitedNode.size()];
			int counter = 0;
			for (int n : notVisitedNode)
			{
				nextNodes[counter] = n;
				costs[counter] = c_ij[currentNode][n];
				counter++;
			}

			/* Order the list of possible next nodes according to its cost */
			int[] orderedNextNodeIndexes = DoubleUtils.sortIndexes(costs, OrderingType.ASCENDING);

			/* Compute the number of elements in the restricted next node list */
			final double minCost = costs[orderedNextNodeIndexes[0]];
			final double maxCost = costs[orderedNextNodeIndexes[orderedNextNodeIndexes.length - 1]];
			final double thresholdCost = minCost + alpha * (maxCost - minCost);
			int numberLinksInRCL = 0;
			for (int index : orderedNextNodeIndexes)
			{
				final int nextNode = nextNodes[index];
				if (c_ij[currentNode][nextNode] > thresholdCost) break;

				numberLinksInRCL++;
			}

			/* The next node is chosen randomly in the restricted next node list */
			final int nextNode = nextNodes[r.nextInt(numberLinksInRCL)];
			nodeSequence.add(nextNode);
			notVisitedNode.remove(nextNode);
			nodeSequenceCost += c_ij[currentNode][nextNode];
		}
		
		nodeSequenceCost += c_ij[nodeSequence.get(nodeSequence.size() - 1)][nodeSequence.get(0)];

		return Pair.of(nodeSequence, nodeSequenceCost);
	}

	/**
	 * This function receives a solution (node sequence and cost), and checks it 
	 * is a valid ring, and that the cost is well calculated.
	 * 
	 * @since 1.0
	 */
	private void checkRing(Pair<ArrayList<Integer>, Double> sol, double[][] c_ij, String message)
	{
		final int N = c_ij.length;
		boolean[] inRing = new boolean[N];
		final ArrayList<Integer> sol_ns = sol.getFirst();
		double checkCost = 0;
		if (sol_ns.size() != N) throw new RuntimeException("The solution created by the greed step is not a ring");

		for (int cont = 0; cont < N; cont++)
		{
			int a_e = sol_ns.get(cont);
			int b_e = (cont == N - 1) ? sol_ns.get(0) : sol_ns.get(cont + 1);
			if (inRing[a_e] == true)
			{
				throw new Net2PlanException("The solution created by the greed step is not a ring");
			}
			else
			{
				checkCost += c_ij[a_e][b_e];
				inRing[a_e] = true;
			}
		}
		if (Math.abs(checkCost - sol.getSecond()) > 1e-3)
		{
			throw new RuntimeException("The solution cost is not well calculated. Solution true cost: " + checkCost + ", solution advertised cost: " + sol.getSecond());
		}
		System.out.println(message + " -- Cost: " + checkCost + ": Sequence of nodes: " + sol.getFirst());
	}


	/**
	 * This function starts from an initial solution. Applies a local search with 
	 * first-fit. Two neighbor solutions are those which have all the links in 
	 * common but two.
	 * 
	 * @since 1.0
	 */
	private Pair<ArrayList<Integer>, Double> computeLocalSearchStep(Pair<ArrayList<Integer>, Double> initialSolution, double[][] c_ij)
	{
		ArrayList<Integer> nodeSequence = initialSolution.getFirst();
		double nodeSequenceCost = initialSolution.getSecond();
		final int N = c_ij.length;
		boolean costImprovementThisIteration = true;
		while (costImprovementThisIteration)
		{
			/* Generate the neighbors of current solution */
			costImprovementThisIteration = false;
			for (int e1 = 0; e1 < N - 2; e1++)
			{
				if (costImprovementThisIteration) break;

				final int e1_a = nodeSequence.get(e1);
				final int e1_b = nodeSequence.get(e1 + 1);
				for (int e2 = e1 + 2; e2 < N; e2++)
				{
					final int e2_a = nodeSequence.get(e2);
					final int e2_b = (e2 == N - 1) ? nodeSequence.get(0) : nodeSequence.get(e2 + 1);
					if ((e1_b == e2_a) || (e2_b == e1_a)) continue;
					if ((e1_b == e2_b) || (e1_a == e2_a)) throw new RuntimeException("Bad");

					final double candidateCost = nodeSequenceCost + (c_ij[e1_a][e2_a] + c_ij[e1_b][e2_b] - c_ij[e1_a][e1_b] - c_ij[e2_a][e2_b]);
					if (candidateCost < nodeSequenceCost)
					{
						/* improvement */
						nodeSequenceCost = candidateCost;
						costImprovementThisIteration = true;
						ArrayList<Integer> oldNodeSequence = (ArrayList<Integer>) nodeSequence.clone();
						for (int cont = 0; cont < e2 - e1; cont++)
							nodeSequence.set(e1 + 1 + cont, oldNodeSequence.get(e2 - cont));
						nodeSequence.set((e2 + 1 == N) ? 0 : e2 + 1, e2_b);
						break;
					}
				}
			}

		}
		return Pair.of(nodeSequence, nodeSequenceCost);
	}
}
